from enum import Enum
SIM_HOST = "https://simulator.home-connect.com"
API_HOST = "https://api.home-connect.com"
ENDPOINT_AUTHORIZE = "/security/oauth/authorize"
ENDPOINT_TOKEN = "/security/oauth/token"
DEFAULT_SCOPES = [ 'IdentifyAppliance', 'Monitor', 'Control', 'Settings'  ]

class Events(str,Enum):
    """ Enum for special event types """
    DATA_REFRESHED = "DATA_REFRESHED"
    CONNECTION_CHANGED = "CONNECTION_CHANGED"
    DEFAULT = "DEFAULT"


EVENT_DATA_REFRESHED = "DATA_REFRESHED"
EVENT_CONNECTION_CHANGED = "CONNECTION_CHANGED"



from __future__ import annotations
from dataclasses import dataclass
import logging
import asyncio
from collections.abc import Callable
from aiohttp import ClientResponse

from .auth import AbstractAuth
from .common import HomeConnectError, DeviceOfflineError
from .const import *

_LOGGER = logging.getLogger(__name__)


class HomeConnectApi():
    """ A class that provides basic API calling facilities to the Home Connect API """
    @dataclass
    class ApiResponse():
        """ Class to encapsulate a service response """
        response:ClientResponse
        status:int
        json:str
        data:any
        error:any

        def __init__(self, response:ClientResponse, json_body):
            self.response = response
            self.status = response.status
            self.json_body = json_body
            self.data = json_body['data'] if json_body and 'data' in json_body else None
            self.error = json_body['error'] if json_body and 'error' in json_body else None

        @property
        def error_key(self) -> str | None:
            """ Dynamically extract the error key from the response """
            if self.error and "key" in self.error:
                return self.error["key"]
            return None

        @property
        def error_description(self) -> str | None:
            """ Dynamically extract the error description from the response """
            if self.error and "description" in self.error:
                return self.error["description"]
            return None


    def __init__(self, auth:AbstractAuth):
        self._auth = auth


    async def _async_request(self, method, endpoint, data=None) -> ApiResponse:
        """ Main function to call the Home Connect API over HTTPS """
        retry = 3
        response = None
        while retry:
            try:
                response = await self._auth.request(method, endpoint, data=data)
                if response.status == 429:    # Too Many Requests
                    wait_time = response.headers.get('Retry-After')
                    _LOGGER.debug('HTTP Error 429 - Too Many Requests. Sleeping for %s seconds and will retry', wait_time)
                    await asyncio.sleep(int(wait_time)+1)
                elif response.status == 401 or response.status >= 500: # Unauthorized
                    # This is probably caused by an expires token so the next retry will get a new one automatically
                    pass
                elif not response.content_length:
                    result = self.ApiResponse(response, None)
                    return result
                else:
                    result = self.ApiResponse(response, await response.json())
                    if result.error_key == "SDK.Error.HomeAppliance.Connection.Initialization.Failed":
                        raise DeviceOfflineError(result.error_description, response=result)
                    return result
            except Exception as ex:
                _LOGGER.debug("Unexpected exeption when calling HomeConnect service", exc_info=ex)
                if not retry:
                    raise HomeConnectError("Unexpected exception when calling HomeConnect service", code=901, inner_exception=ex) from ex
            finally:
                if response:
                    response.close()
                    response = None
            retry -= 1

        # all retries were exhausted without a valid response
        raise HomeConnectError("Failed to get a valid response from Home Connect server", 902)


    async def async_get(self, endpoint, lang='en-GB') -> ApiResponse:
        """ Implements a HTTP GET request """
        return await self._async_request('get', endpoint)

    async def async_put(self, endpoint:str, data:str, lang='en-GB') -> ApiResponse:
        """ Implements a HTTP PUT request """
        return await self._async_request('put', endpoint, data=data)

    async def async_delete(self, endpoint:str) -> ApiResponse:
        """ Implements a HTTP DELETE request """
        return await self._async_request('delete', endpoint)

    async def async_get_event_stream(self, endpoint):
        """ Returns a Server Sent Events (SSE) stream to be consumed by the caller """
        return await self._auth.stream(endpoint)


    async def async_stream(self, endpoint:str, event_handler:Callable[[str], None]):
        """ Implements a SSE consumer which calls the defined event handler on every new event"""
        backoff = 2
        event_source = None
        while True:
            try:
                event_source = await self._auth.stream(endpoint)
                await event_source.connect()
                async for event in event_source:
                    backoff = 1
                    try:
                        await event_handler(event)
                    except Exception as ex:
                        _LOGGER.exception('Unhandled exception in stream event handler', exc_info=ex)
            except asyncio.CancelledError:
                break
            except ConnectionRefusedError as ex:
                _LOGGER.exception('ConnectionRefusedError in SSE connection refused. Will try again', exc_info=ex)
            except ConnectionError as ex:
                error_code = self.parse_sse_error(ex.args[0])
                if error_code == 429:
                    backoff *= 2
                    if backoff > 3600: backoff = 3600
                    elif backoff < 60: backoff = 60
                    _LOGGER.info('Got error 429 when opening event stream connection, will sleep for %s seconds and retry', backoff)
                else:
                    _LOGGER.exception('ConnectionError in SSE event stream. Will wait for %d seconds and retry ', backoff, exc_info=ex)
                    backoff *= 2
                    if backoff > 120: backoff = 120

                await asyncio.sleep(backoff)

            except asyncio.TimeoutError:
                # it is expected that the connection will time out every hour
                _LOGGER.debug("The SSE connection timeout, will renew and retry")
            except Exception as ex:
                _LOGGER.exception('Exception in SSE event stream. Will wait for %d seconds and retry ', backoff, exc_info=ex)
                await asyncio.sleep(backoff)
                backoff *= 2
                if backoff > 120: backoff = 120

            finally:
                if event_source:
                    await event_source.close()
                    event_source = None


    def parse_sse_error(self, error:str) -> int:
        """ Helper function to parse the error code from a SSE exception """
        try:
            parts = error.split(': ')
            error_code = int(parts[-1])
            return error_code
        except:
            return 0



from .homeconnect import HomeConnect
from .appliance import Appliance
from .auth import AuthManager, AbstractAuth
from .common import HomeConnectError
#from .const import EVENT_DATA_REFRESHED, EVENT_CONNECTION_CHANGED

EVENT_DATA_REFRESHED = "DATA_REFRESHED"
EVENT_CONNECTION_CHANGED = "CONNECTION_CHANGED"



from __future__ import annotations
import fnmatch
import inspect
import re
from typing import Callable
from collections.abc import Sequence

from .appliance import Appliance


WILDCARD_KEY = "WILDCARD"
DEFAULT_KEY = "DEFAULT"
class CallbackRegistry():
    """ Calss for managing callback registration and notifications """

    def __init__(self) -> None:
        self._callbacks = {}

    def register_callback(self, callback:Callable[[Appliance, str], None] | Callable[[Appliance, str, any], None], keys:str|Sequence[str], appliance:Appliance|str = None):
        """ Register callback for change event notifications

        Use the Appliance.register_callback() to register for appliance data update events
        """

        if not isinstance(keys, list):
            keys = [ keys ]

        haid = appliance.haId if isinstance(appliance, Appliance) else appliance

        if haid not in self._callbacks:
            self._callbacks[haid] = {}

        for key in keys:
            if '*' in key:
                callback_record = {
                    "key": re.compile(fnmatch.translate(key), re.IGNORECASE),
                    "callback": callback
                }
                if WILDCARD_KEY not in self._callbacks[haid]:
                    self._callbacks[haid][WILDCARD_KEY] = []
                self._callbacks[haid][WILDCARD_KEY].append(callback_record)
            else:
                if key not in self._callbacks[haid]:
                    self._callbacks[haid][key] = set()
                self._callbacks[haid][key].add(callback)

    def deregister_callback(self,
        callback:Callable[[Appliance], None] | Callable[[Appliance, str], None] | Callable[[Appliance, str, any], None],
        keys:str|Sequence[str],
        appliance:Appliance|str = None
    ):
        """ Clear a callback that was prevesiously registered so it stops getting notifications """

        if not isinstance(keys, list):
            keys = [ keys ]

        haid = appliance.haId if isinstance(appliance, Appliance) else appliance

        if haid not in self._callbacks:
            self._callbacks[haid] = {}

        for key in keys:
            if '*' in key:
                callback_record = {
                    "key": re.compile(fnmatch.translate(key), re.IGNORECASE),
                    "callback": callback
                }
                try:
                    if haid in self._callbacks and WILDCARD_KEY in self._callbacks[haid]:
                        self._callbacks[haid][WILDCARD_KEY].remove(callback_record)
                except ValueError:
                    # ignore if the value is not found in the list
                    pass
            else:
                if haid in self._callbacks and key in self._callbacks[haid]:
                    self._callbacks[haid][key].remove(callback)

    def clear_all_callbacks(self):
        """ Clear all the registered callbacks """
        self._callbacks = {}

    def clear_appliance_callbacks(self, appliance:Appliance|str):
        """ Clear all the registered callbacks """
        haid = appliance.haId if isinstance(appliance, Appliance) else appliance

        if haid in self._callbacks:
            del self._callbacks[haid]

    async def async_broadcast_event(self, appliance:Appliance, event_key:str, value:any = None) -> None:
        """ Broadcast an event to all subscribed callbacks """

        handled:bool = False
        haid = appliance.haId

        # dispatch simple event callbacks
        for haid in [ handler for handler in [None, appliance.haId] if handler in self._callbacks] :
            if event_key in self._callbacks[haid]:
                for callback in self._callbacks[haid][event_key]:
                    await self._async_call(callback, appliance, event_key, value)
                    handled = True

            # dispatch wildcard or value based callbacks
            if WILDCARD_KEY in self._callbacks[haid]:
                for callback_record in self._callbacks[haid][WILDCARD_KEY]:
                    if callback_record["key"].fullmatch(event_key):
                        callback = callback_record['callback']
                        await self._async_call(callback, appliance, event_key, value)
                        handled = True

            # dispatch default callbacks for unhandled events
            if not handled and DEFAULT_KEY in self._callbacks[haid]:
                for callback in self._callbacks[DEFAULT_KEY]:
                    self._async_call(callback, appliance, event_key, value)


    async def _async_call(self, callback:Callable, appliance:Appliance, event_key:str, value:any) -> None:
        """ Helper funtion to make the right kind of call to the callback funtion """
        sig = inspect.signature(callback)
        param_count = len(sig.parameters)
        if inspect.iscoroutinefunction(callback):
            if param_count == 3:
                await callback(appliance, event_key, value)
            elif param_count == 2:
                await callback(appliance, event_key)
            elif param_count == 1:
                await callback(appliance)
            elif param_count == 0:
                await callback()
            else:
                raise ValueError(f"Unexpected number of callback parameters: {sig}")
        else:
            if param_count == 3:
                callback(appliance, event_key, value)
            elif param_count == 2:
                callback(appliance, event_key)
            elif param_count == 1:
                callback(appliance)
            elif param_count == 0:
                callback()
            else:
                raise ValueError(f"Unexpected number of callback parameters: {sig}")
